import numpy as np


class Affine(np.ndarray):
    """
    Class for general (d+1)x(d+1) affine matrices, and in particular d=3 (3d space)
    Usage examples:
    >>> affine1 = Affine(np.eye(4))
    >>> affine2 = Affine(direction=np.eye(3), spacing=[0.33, 1, 0.33], origin=[-90.3, 10, 1.44])
    >>> index = [4, 0, 9]
    >>> coord = affine2.index2coord(index)
    >>> print(coord)
    [-88.98  10.     4.41]
    """

    # keys for the origin and M matrix parts in the affine matrix
    _origin_key = (slice(-1), -1)
    _m_key = (slice(-1), slice(-1))

    def __new__(cls, affine=None, *, direction=None, spacing=None, origin=None):
        """
        Construct a numpy array of class Affine. Initialize Affine in one of the following ways:
        1. (d+1)x(d+1) matrix as affine (d is the dimension of the space)
        2. affine=None and construction from direction, spacing and origin parameters
        :param affine: (d+1)x(d+1) affine matrix, comprised of the M matrix and origin shift b: y = M*x + b
        x is the index vector of length d and y is the corresponding physical coordinates vector of the same length
        :param direction: dxd direction matrix (only rotations without scaling)
        :param spacing: scaling of the axes - vector of length d
        :param origin: the origin - b in the formula above - vector of length d (or a scalar)
        :return: numpy.ndarray viewed as type 'Affine'
        """
        if affine is None:
            affine = cls.construct_affine(direction, spacing, origin)
        obj = np.asarray(affine).view(cls)  # return array view of type Affine
        return obj
    
    def __init__(self, affine=None, *, direction=None, spacing=None, origin=None):
        self.dim = self.shape[0] - 1
        if affine is None:
            self._spacing = np.asarray(spacing)
            self._direction = np.asarray(direction)
        else:
            # TODO: reconsider calculating it here
            self._spacing = self.affine2spacing(self)
            self._direction = self.affine2direction(self, self.spacing)

    def index2coord(self, index_vector):
        """Return y according to y = M*x + b"""
        return self._m_matrix @ index_vector + self.origin

    def __matmul__(self, other):
        return super().__matmul__(other).view(np.ndarray)

    def __getitem__(self, item):
        return super().__getitem__(item).view(np.ndarray)

    def clone(self):
        return Affine(self.copy())

    # Affine properties in addition to the numpy array
    @property
    def origin(self):
        return self[self._origin_key]

    @origin.setter
    def origin(self, value):
        self[self._origin_key] = value

    @property
    def spacing(self):
        return self._spacing

    @spacing.setter
    def spacing(self, value):
        value = np.asarray(value)
        self._m_matrix = self._m_matrix @ np.diag(value / self._spacing)
        self._spacing = np.abs(value)  # the spacing must be positive (or at least nonnegative)

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, value):
        value = np.asarray(value)
        self._m_matrix = value @ np.diag(self.spacing)
        self._direction = value

    # Internal property - m matrix
    @property
    def _m_matrix(self):
        return self[self._m_key]

    @_m_matrix.setter
    def _m_matrix(self, value):
        self[self._m_key] = value

    # Static methods for affine construction and components
    @staticmethod
    def construct_affine(direction, spacing, origin):
        direction = np.asarray(direction)
        dim = direction.shape[0]
        affine = np.eye(dim + 1)
        affine[Affine._m_key] = direction @ np.diag(spacing)
        affine[Affine._origin_key] = origin
        return affine

    @staticmethod
    def affine2origin(affine):
        return affine[Affine._origin_key]

    @staticmethod
    def affine2spacing(affine):
        dim = affine.shape[0] - 1
        return np.linalg.norm(affine[Affine._m_key] @ np.eye(dim), axis=0)

    @staticmethod
    def affine2direction(affine, spacing=None):
        if spacing is None:
            spacing = Affine.affine2spacing(affine)
        return affine[Affine._m_key] @ np.diag(1 / spacing)

    @staticmethod
    def affine2comps(affine, spacing=None):
        if spacing is None:
            spacing = Affine.affine2spacing(affine)
        return Affine.affine2direction(affine, spacing), spacing, Affine.affine2origin(affine)



from typing import Union

import nibabel as nib
import numpy as np

from medio.metadata.affine import Affine
from medio.metadata.metadata import MetaData


class NibIO:
    coord_sys = 'nib'
    RGB_DTYPE = np.dtype([('R', np.uint8),
                          ('G', np.uint8),
                          ('B', np.uint8)])
    RGBA_DTYPE = np.dtype([('R', np.uint8),
                           ('G', np.uint8),
                           ('B', np.uint8),
                           ('A', np.uint8)])

    @staticmethod
    def read_img(input_path, desired_axcodes=None, header=False, channels_axis=None):
        """
        Reads a NIFTI file and returns the image array and metadata
        :param input_path: path-like (str or pathlib.Path) of the image file to read
        :param desired_axcodes: str, tuple of str or None - the desired orientation of the image to be returned
        :param channels_axis: if not None and the array dtype is structured, stacks the channels along channels_axis
        :param header: whether to include a header attribute with additional metadata in the returned metadata
        :return: image array and corresponding metadata
        """
        img_struct = nib.load(input_path)
        orig_ornt_str = ''.join(nib.aff2axcodes(img_struct.affine))
        if desired_axcodes is not None:
            img_struct = NibIO.reorient(img_struct, desired_axcodes)
        img = np.asanyarray(img_struct.dataobj)
        if channels_axis is not None:
            img = NibIO.unravel_array(img, channels_axis)
        affine = Affine(img_struct.affine)
        metadata = MetaData(affine=affine, orig_ornt=orig_ornt_str, coord_sys=NibIO.coord_sys)
        if header:
            metadata.header = {key: img_struct.header[key] for key in img_struct.header.keys()}
        return img, metadata

    @staticmethod
    def save_img(filename, img, metadata, use_original_ornt=True, channels_axis=None):
        """
        Saves the given image as a NIFTI file.
        :param filename: path-like output filename, including a '.nii.gz' or '.nii' suffix
        :param img: image data array
        :param metadata: the matching metadata
        :param use_original_ornt: whether to use the original orientation of the image of not
        :param channels_axis: if not None gives the channels axis of img (for channeled images RGB/RGBA)
        """
        if channels_axis is not None:
            img = NibIO.pack_channeled_img(img, channels_axis)
        orig_coord_sys = metadata.coord_sys
        metadata.convert(NibIO.coord_sys)
        img_struct = nib.Nifti1Image(img, metadata.affine)
        desired_axcodes = metadata.orig_ornt if use_original_ornt else None
        metadata.convert(orig_coord_sys)
        img_struct = NibIO.reorient(img_struct, desired_axcodes)
        nib.save(img_struct, filename)

    @staticmethod
    def reorient(img_struct, desired_axcodes: Union[tuple, str, None]):
        """Reorient a nibabel image to a desired orientation described by desired_axcodes strings tuple, for example
        ('L', 'P', 'I'). If desired_axcodes is None it returns the given img_struct"""
        if desired_axcodes is not None:
            if isinstance(desired_axcodes, str):
                desired_axcodes = tuple(desired_axcodes)
            start_ornt = nib.orientations.io_orientation(img_struct.affine)
            end_ornt = nib.orientations.axcodes2ornt(desired_axcodes)
            ornt_tform = nib.orientations.ornt_transform(start_ornt, end_ornt)
            img_struct = img_struct.as_reoriented(ornt_tform)
        return img_struct

    @staticmethod
    def unravel_array(array, channels_axis=-1):
        """Simplify array dtype if it is a structured data type. For example, if the array if of RGB dtype:
        np.dtype([('R', 'uint8'), ('G', 'uint8'), ('B', 'uint8')])
        Convert it into an array with dtype 'uint8' and 3 channels for RGB in an additional last dimension"""
        dtype = array.dtype
        if not (hasattr(dtype, '__len__') and len(dtype) > 1):
            return array
        return np.stack([array[field] for field in dtype.fields], axis=channels_axis)

    @staticmethod
    def pack_channeled_img(img, channels_axis):
        dtype = img.dtype
        if not np.issubdtype(dtype, np.uint8):
            raise ValueError(f'RGB or RGBA images must have dtype "np.uint8", got: "{dtype}"')
        n_channels = img.shape[channels_axis]
        img = np.moveaxis(img, channels_axis, -1)
        r_channel = img[..., 0]
        if n_channels == 3:
            img_rgb = np.empty_like(r_channel, dtype=NibIO.RGB_DTYPE)
            img_rgb['R'] = r_channel
            img_rgb['G'] = img[..., 1]
            img_rgb['B'] = img[..., 2]
            return img_rgb
        elif n_channels == 4:
            img_rgba = np.empty_like(r_channel, dtype=NibIO.RGBA_DTYPE)
            img_rgba['R'] = r_channel
            img_rgba['G'] = img[..., 1]
            img_rgba['B'] = img[..., 2]
            img_rgba['A'] = img[..., 3]
            return img_rgba
        else:
            raise ValueError(f'Invalid number of channels: {n_channels}, should be 3 (RGB) or 4 (RGBA)')



import pprint
from copy import deepcopy

import numpy as np
from nibabel import aff2axcodes

from medio.metadata.affine import Affine
from medio.metadata.convert_nib_itk import convert_nib_itk, inv_axcodes, convert_affine


class MetaData:
    def __init__(self, affine, orig_ornt=None, coord_sys='itk', header=None):
        """
        Initialize medical image's metadata
        :param affine: affine matrix of class Affine, numpy float array of shape (4, 4)
        :param orig_ornt: orientation string code, str of length 3 or None (was not computed because the image was not
        reoriented)
        :param coord_sys: 'itk' or 'nib', the coordinate system of the given affine and orientation: itk or nib (nifti)
        :param header: additional metadata dictionary with string keys
        """
        if not isinstance(affine, Affine):
            affine = Affine(affine)
        self.affine = affine
        self.orig_ornt = orig_ornt
        self._ornt = None
        self.check_valid_coord_sys(coord_sys)
        self.coord_sys = coord_sys
        self.header = header

    @staticmethod
    def check_valid_coord_sys(coord_sys):
        if coord_sys not in ('itk', 'nib'):
            raise ValueError('Metadata coord_sys must be "itk" or "nib"')

    def __repr__(self):
        sep = ' ' if self.header is None else '\n'
        return (f'Affine:\n'
                f'{self.affine}\n'
                f'Spacing: {self.spacing}\n'
                f'Coordinate system: {self.coord_sys}\n'
                f'Orientation: {self.ornt}\n'
                f'Original orientation: {self.orig_ornt}\n'
                f'Header:{sep}'
                f'{pprint.pformat(self.header, indent=4)}')

    def convert(self, dest_coord_sys):
        """
        Converts the metadata coordinate system inplace to dest_coord_sys. Affects affine, ornt and orig_ornt
        :param dest_coord_sys: the destination coordinate system - 'itk' or 'nib' (nifti)
        """
        self.check_valid_coord_sys(dest_coord_sys)
        if dest_coord_sys != self.coord_sys:
            self.affine, self._ornt, self.orig_ornt = convert_nib_itk(self.affine, self._ornt, self.orig_ornt)
            self.coord_sys = dest_coord_sys

    def clone(self):
        return MetaData(affine=self.affine.clone(), orig_ornt=self.orig_ornt, coord_sys=self.coord_sys,
                        header=deepcopy(self.header))

    def get_ornt(self):
        """Returns current orientation based on the affine and coordinate system"""
        if self.coord_sys == 'nib':
            ornt_tup = aff2axcodes(self.affine)
        elif self.coord_sys == 'itk':
            ornt_tup = inv_axcodes(aff2axcodes(convert_affine(self.affine)))
        else:
            raise ValueError(f'Invalid coord_sys: "{self.coord_sys}"')
        ornt_str = ''.join(ornt_tup)
        return ornt_str

    @property
    def ornt(self):
        if self._ornt is None:
            self._ornt = self.get_ornt()
            # if self.orig_ornt is also None, the affine was not reoriented and the original orientation is the same
            if self.orig_ornt is None:
                self.orig_ornt = self._ornt
        return self._ornt

    @property
    def spacing(self):
        return self.affine.spacing

    def is_right_handed_ornt(self):
        """Check whether the affine orientation is right or left handed. The sign of the triple product of the
        direction matrix is calculated with a determinant. It should be +1 or -1 because it is a rotation matrix.
        +1 (-1) indicates right (left) handed orientation.
        To be used primarily before saving a dicom file or series"""
        if self.affine.dim != 3:
            raise ValueError('Right handed orientation is relevant only to a 3d space')
        return np.linalg.det(self.affine.direction) > 0


def is_right_handed_axcodes(axcodes):
    letter_vec_dict = {'R': [1, 0, 0],
                       'L': [-1, 0, 0],
                       'A': [0, 1, 0],
                       'P': [0, -1, 0],
                       'I': [0, 0, 1],
                       'S': [0, 0, -1]
                       }
    u, v, n = [letter_vec_dict[letter] for letter in axcodes]
    ornt_sign = np.dot(np.cross(u, v), n)
    if ornt_sign not in (-1, 1):
        raise ValueError(f'Invalid axcodes: "{axcodes}"')
    return ornt_sign == 1


def flip_last_axcodes(axcodes):
    return axcodes[:-1] + inv_axcodes(axcodes[-1])


def check_dcm_ornt(desired_ornt, metadata, allow_dcm_reorient=False):
    """Check whether the orientation desired_ornt is right handed before saving image as a dicom
    :param desired_ornt: the desired orientation for the saver
    :param metadata: if desired_ornt is None (not set), use metadata.ornt
    :param allow_dcm_reorient: whether to allow automatic reorientation to a right handed orientation or not
    :return: right handed desired_ornt or ValueError
    """
    # first set the desired orientation
    if desired_ornt is None:
        desired_ornt = metadata.ornt
    if is_right_handed_axcodes(desired_ornt):
        return desired_ornt
    else:
        right_handed_ornt = flip_last_axcodes(desired_ornt)
        if allow_dcm_reorient:
            return right_handed_ornt
        else:
            raise ValueError(f'The desired orientation "{desired_ornt}" is left handed, whereas saving dicom is '
                             f'possible only with a right handed orientation. \nYou can either pass the saver '
                             f'parameter allow_dcm_reorient=True to allow automatic reorientation (in this case to '
                             f'"{right_handed_ornt}"), or \nreorient yourself before saving the image as a dicom.')



from pathlib import Path

import pydicom
import numpy as np
from dicom_numpy import combine_slices

from medio.backends.pdcm_unpack_ds import unpack_dataset
from medio.metadata.metadata import MetaData
from medio.metadata.pdcm_ds import convert_ds, MultiFrameFileDataset


class PdcmIO:
    coord_sys = 'itk'
    DEFAULT_CHANNELS_AXIS = 0  # in the transposed image

    @staticmethod
    def read_img(input_path, header=False, channels_axis=None, globber='*', allow_default_affine=False):
        """
        Read a dicom file or folder (series) and return the numpy array and the corresponding metadata
        :param input_path: path-like object (str or pathlib.Path) of the file or directory to read
        :param globber: relevant for a directory - globber for selecting the series files (all files by default)
        :param header: whether to include a header attribute with additional metadata in the returned metadata (single
        file only)
        :param channels_axis: if not None and the image is channeled (e.g. RGB) move the channels to channels_axis in
        the returned image array
        :param allow_default_affine: whether to allow default affine when some tags are missing (multiframe file only)
        :return: numpy array and metadata
        """
        input_path = Path(input_path)
        if input_path.is_dir():
            return PdcmIO.read_dcm_dir(input_path, globber, channels_axis=channels_axis)
        else:
            return PdcmIO.read_dcm_file(input_path, header, allow_default_affine=allow_default_affine,
                                        channels_axis=channels_axis)

    @staticmethod
    def read_dcm_file(filename, header=False, allow_default_affine=False, channels_axis=None):
        """Read a single dicom file"""
        ds = pydicom.dcmread(filename)
        ds = convert_ds(ds)
        if ds.__class__ is MultiFrameFileDataset:
            img, affine = unpack_dataset(ds, allow_default_affine=allow_default_affine)
        else:
            img, affine = combine_slices([ds])
        metadata = PdcmIO.aff2meta(affine)
        if header:
            metadata.header = {str(key): ds[key] for key in ds.keys()}
        img = PdcmIO.move_channels_axis(img, samples_per_pixel=ds.SamplesPerPixel, channels_axis=channels_axis)
        return img, metadata

    @staticmethod
    def read_dcm_dir(input_dir, globber='*', channels_axis=None):
        """Reads a 3D dicom image: input path can be a file or directory (DICOM series)"""
        # find all dicom files within the specified folder, read every file separately and sort them by InstanceNumber
        files = list(Path(input_dir).glob(globber))
        if len(files) == 0:
            raise FileNotFoundError(f'Received an empty directory: "{input_dir}"')
        slices = [pydicom.dcmread(filename) for filename in files]
        slices.sort(key=lambda ds: ds.get('InstanceNumber', 0))
        img, affine = combine_slices(slices)
        metadata = PdcmIO.aff2meta(affine)
        img = PdcmIO.move_channels_axis(img, samples_per_pixel=slices[0].SamplesPerPixel, channels_axis=channels_axis)
        return img, metadata

    @staticmethod
    def aff2meta(affine):
        return MetaData(affine, coord_sys=PdcmIO.coord_sys)

    @staticmethod
    def move_channels_axis(array, samples_per_pixel, source=DEFAULT_CHANNELS_AXIS, channels_axis=None):
        # TODO: the following assert is not always True, depends on Planar Configuration (0028,0006)
        # assert array.shape[source] == samples_per_pixel
        if (samples_per_pixel > 1) and (channels_axis is not None):
            return np.moveaxis(array, source, channels_axis)
        return array

    @staticmethod
    def save_arr2dcm_file(output_filename, template_filename, img_arr, dtype=None, keep_rescale=False):
        """
        Writes a dicom single file image using template file, without the intensity transformation from template dataset
        unless keep_rescale is True
        :param output_filename: path-like object of the output file to be saved
        :param template_filename: the single dicom scan whose metadata is used
        :param img_arr: numpy array of the image to be saved, should be in the same orientation as template_filename
        :param dtype: the dtype for the numpy array, for example 'int16'. If None - will use the dtype of the template
        :param keep_rescale: whether to keep intensity rescale values
        """
        ds = pydicom.dcmread(template_filename)
        ds = convert_ds(ds)
        if not keep_rescale:
            if isinstance(ds, MultiFrameFileDataset):
                ds.del_intensity_trans()
            else:
                del ds.RescaleSlope
                del ds.RescaleIntercept
        if dtype is None:
            img_arr = img_arr.astype(ds.pixel_array.dtype, copy=False)
        else:
            img_arr = img_arr.astype(dtype, copy=False)
        ds.PixelData = img_arr.tobytes()
        ds.save_as(output_filename)



"""
This module is equivalent to dicom_numpy's module: combine_slices.py, but here for a single dicom dataset
"""
import logging

import numpy as np
from dicom_numpy.combine_slices import _validate_image_orientation, _extract_cosines, _requires_rescaling

logger = logging.getLogger(__name__)


def unpack_dataset(dataset, rescale=None, allow_default_affine=False):
    """
    Given a pydicom dataset of a single image file return three-dimensional numpy array.
    Also calculate a 4x4 affine transformation matrix that converts the ijk-pixel-indices
    into the xyz-coordinates in the DICOM patient's coordinate system.

    Returns a two-tuple containing the 3D-ndarray and the affine matrix.

    If `rescale` is set to `None` (the default), then the image array dtype
    will be preserved, unless any of the DICOM images contain either the
    `Rescale Slope
    <https://dicom.innolitics.com/ciods/ct-image/ct-image/00281053>`_ or the
    `Rescale Intercept <https://dicom.innolitics.com/ciods/ct-image/ct-image/00281052>`_
    attributes.  If either of these attributes are present they will be applied.

    If `rescale` is `True` the voxels will be cast to `float32`, if set to
    `False`, the original dtype will be preserved even if DICOM rescaling information is present.

    The returned array has the column-major byte-order.

    This function requires that the datasets:

    - Be in same series (have the same
      `Series Instance UID <https://dicom.innolitics.com/ciods/ct-image/general-series/0020000e>`_,
      `Modality <https://dicom.innolitics.com/ciods/ct-image/general-series/00080060>`_,
      and `SOP Class UID <https://dicom.innolitics.com/ciods/ct-image/sop-common/00080016>`_).
    - The binary storage of each slice must be the same (have the same
      `Bits Allocated <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280100>`_,
      `Bits Stored <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280101>`_,
      `High Bit <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280102>`_, and
      `Pixel Representation <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280103>`_).
    - The image slice must approximately form a grid. This means there can not
      be any missing internal slices (missing slices on the ends of the dataset
      are not detected).
    - It also means that  each slice must have the same
      `Rows <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280010>`_,
      `Columns <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280011>`_,
      `Pixel Spacing <https://dicom.innolitics.com/ciods/ct-image/image-plane/00280030>`_, and
      `Image Orientation (Patient) <https://dicom.innolitics.com/ciods/ct-image/image-plane/00200037>`_
      attribute values.
    - The direction cosines derived from the
      `Image Orientation (Patient) <https://dicom.innolitics.com/ciods/ct-image/image-plane/00200037>`_
      attribute must, within 1e-4, have a magnitude of 1.  The cosines must
      also be approximately perpendicular (their dot-product must be within
      1e-4 of 0).  Warnings are displayed if any of these approximations are
      below 1e-8, however, since we have seen real datasets with values up to
      1e-4, we let them pass.
    - The `Image Position (Patient) <https://dicom.innolitics.com/ciods/ct-image/image-plane/00200032>`_
      values must approximately form a line.

    If any of these conditions are not met, a `dicom_numpy.DicomImportException` is raised.
    """
    try:
        _validate_image_orientation(dataset.ImageOrientationPatient)
        transform = _ijk_to_patient_xyz_transform_matrix(dataset)
    except AttributeError as e:
        if allow_default_affine:
            transform = np.eye(4)
        else:
            raise AttributeError(str(e) + "\nTry using: allow_default_affine=True")

    voxels = _unpack_pixel_array(dataset, rescale)
    return voxels, transform


def _unpack_pixel_array(dataset, rescale=None):
    voxels = dataset.pixel_array.T

    if rescale is None:
        rescale = _requires_rescaling(dataset)

    if rescale:
        voxels = voxels.astype('int16', copy=False)  # TODO: it takes time! Consider view.
        slope = getattr(dataset, 'RescaleSlope', 1)
        intercept = getattr(dataset, 'RescaleIntercept', 0)
        if int(slope) == slope and int(intercept) == intercept:
            slope = int(slope)
            intercept = int(intercept)
        voxels = voxels * slope + intercept

    return voxels


def _ijk_to_patient_xyz_transform_matrix(dataset):
    image_orientation = dataset.ImageOrientationPatient
    row_cosine, column_cosine, slice_cosine = _extract_cosines(image_orientation)

    row_spacing, column_spacing = dataset.PixelSpacing
    # slice_spacing = dataset.get('SpacingBetweenSlices', 0)

    transform = np.identity(4, dtype=np.float32)

    transform[:3, 0] = row_cosine * column_spacing
    transform[:3, 1] = column_cosine * row_spacing
    transform[:3, 2] = (np.array(dataset.slice_position(-1)) - dataset.slice_position(0)
                        ) / (dataset.NumberOfFrames - 1)
    # transform[:3, 2] = slice_cosine * slice_spacing

    transform[:3, 3] = dataset.ImagePositionPatient

    return transform



from functools import partial

from pydicom.uid import generate_uid

# Given by Medical Connections (http://www.medicalconnections.co.uk/FreeUID.html)
MEDIO_ROOT_UID = '1.2.826.0.1.3680043.10.513.'


generate_uid = partial(generate_uid, prefix=MEDIO_ROOT_UID)



"""
NiBabel <-> ITK orientation and affine conversion utilities.
The conventions of nibabel and itk are different and this module supplies functions which convert between
these conventions.

Orientation
-----------
In nibabel each axis code indicates the ending direction - RAS+: L -> R, P -> A, I -> S
In itk it corresponds to the converse of nibabel - RAS: R -> L, A -> P, S -> I

Affine
------
In itk, the direction matrix (3x3 upper left affine with unit spacings) of RAS orientation image is:
[[1, 0, 0],
 [0, 1, 0],
 [0, 0, -1]]

In nibabel it is LPI+:
[[-1, 0, 0],
 [0, -1, 0],
 [0, 0, -1]]

The matrix convert_aff_mat accounts for this difference (for all possible orientations, not only RAS).

Usage
=====
Works both ways: itk -> nib and nib -> itk, the usage is the same:
>>> new_affine, new_axcodes = convert_nib_itk(affine, axcodes)
"""

import numpy as np

from medio.metadata.affine import Affine
from medio.utils.two_way_dict import TwoWayDict

# store compactly axis directions codes
axes_inv = TwoWayDict()
axes_inv['R'] = 'L'
axes_inv['A'] = 'P'
axes_inv['S'] = 'I'


def inv_axcodes(axcodes):
    """Inverse axes codes chars, for example: SPL -> IAR"""
    if axcodes is None:
        return None
    new_axcodes = ''
    for code in axcodes:
        new_axcodes += axes_inv[code]
    return new_axcodes


def convert_affine(affine):
    # conversion matrix of the affine from itk to nibabel and vice versa
    convert_aff_mat = np.diag([-1, -1, 1, 1])
    # for 2d image:
    if affine.shape[0] == 3:
        convert_aff_mat = np.diag([-1, -1, 1])
    new_affine = convert_aff_mat @ affine
    if isinstance(affine, Affine):
        new_affine = Affine(new_affine)
    return new_affine


def convert_nib_itk(affine, *axcodes):
    """Convert affine and orientations (original and current orientations) from nibabel to itk and vice versa"""
    new_affine = convert_affine(affine)
    new_axcodes = []
    for axcode in axcodes:
        new_axcodes += [inv_axcodes(axcode)]

    return (new_affine, *new_axcodes)



import os

import itk
import numpy as np
import pandas as pd

from medio.tests.itk_dcm_orientations.orientations_utils import (ornt_list, ornt_direction_dict, is_right_handed_ornt,
                                                                 direction2ornt)
from medio.tests.itk_dcm_orientations.itk_utils import set_image_direction, get_image_direction


def get_saved_ornt(image, desired_ornt, remove=True):
    filename = desired_ornt + '.dcm'
    desired_direction = ornt_direction_dict[desired_ornt]
    set_image_direction(image, desired_direction)
    itk.imwrite(image, filename)
    saved_image = itk.imread(filename)
    if remove:
        os.remove(filename)
    saved_direction = get_image_direction(saved_image)
    is_equal = np.array_equal(saved_direction, desired_direction)
    if is_equal:
        saved_ornt = desired_ornt
    else:
        saved_ornt = direction2ornt(saved_direction)
    return saved_ornt, is_equal


arr = np.random.randint(0, 256, size=(10, 30, 25), dtype='uint8')  # (slices, cols, rows)
arr_rgb = np.random.randint(0, 256, size=(10, 30, 25, 3), dtype='uint8')  # (slices, cols, rows, channels)
img = itk.image_from_array(arr)
img_rgb = itk.image_from_array(arr_rgb, is_vector=True)

df = pd.DataFrame(columns=['Orientation',
                           'Right/Left-handed orientation',
                           'Success',
                           'Saved orientation',
                           'RGB success',
                           'RGB saved orientation'
                           ])

for ornt in ornt_list:
    state_dict = dict.fromkeys(df.columns)
    state_dict['Orientation'] = ornt
    state_dict['Right/Left-handed orientation'] = 'R' if is_right_handed_ornt(ornt) else 'L'
    # test 3d dicom
    state_dict['Saved orientation'], state_dict['Success'] = get_saved_ornt(img, ornt)
    # test 3d RGB dicom
    state_dict['RGB saved orientation'], state_dict['RGB success'] = get_saved_ornt(img_rgb, ornt)
    df = df.append(state_dict, ignore_index=True)

df.sort_values('Right/Left-handed orientation', ascending=False, inplace=True)
df.to_csv('itk_dcm_orientations.csv', index=False)

right_handed = np.array(df['Right/Left-handed orientation'] == 'R')
success = np.array(df['Success'])
rgb_success = np.array(df['RGB success'])
rai_ornt = np.array(df['Orientation'] == 'RAI')

print('Right-handed == Success:', np.array_equal(right_handed, success))
print('Right-handed == RGB success:', np.array_equal(right_handed, rgb_success))
print('"RAI" orientation == RGB success:', np.array_equal(rai_ornt, rgb_success))



"""
This module is based on the C++ ITK code - itkSpatialOrientation.h
"""

import itertools
from enum import IntEnum

from medio.utils.two_way_dict import TwoWayDict


class AxCodes(IntEnum):
    UNKNOWN = 0
    R = 2  # Right
    L = 3  # Left
    P = 4  # Posterior
    A = 5  # Anterior
    I = 8  # Inferior
    S = 9  # Superior


class AxMajorness(IntEnum):
    Primary = 0
    Secondary = 8
    Tertiary = 16


class ItkOrientationCode:
    INVALID = AxCodes.UNKNOWN
    # 48 valid orientations are added as attributes


def itk_orientation_code(ax_code):
    """ax_code is string or tuple of valid orientation, e.g. 'LPI', ('A', 'R', 'S')"""
    prime, second, tertiary = [getattr(AxCodes, axis) for axis in ax_code]
    return (prime << AxMajorness.Primary) + (second << AxMajorness.Secondary) + (tertiary << AxMajorness.Tertiary)


# adding all 48 possible orientation codes to ItkOrientationCode class

ax_codes_iter = itertools.chain(
    *map(itertools.permutations,
         itertools.product(('R', 'L'), ('A', 'P'), ('I', 'S'))
         ))

# two way dictionary that translates itk numerical orientation codes to orientation strings and vice versa
codes_str_dict = TwoWayDict()
codes_str_dict[None] = None

for ax_code in ax_codes_iter:
    ax_code_str = ''.join(ax_code)
    code = itk_orientation_code(ax_code)
    setattr(ItkOrientationCode, ax_code_str, code)
    codes_str_dict[ax_code_str] = code



from pathlib import Path


def is_file_suffix(filename, suffixes, check_exist=True):
    """
    is_file + check for suffix
    :param filename: pathlike object
    :param suffixes: tuple of possible suffixes
    :param check_exist: whether to check the file's existence
    :return: bool
    """
    if check_exist and not Path(filename).is_file():
        return False
    return str(filename).endswith(suffixes)


def is_nifti(filename, check_exist=True):
    return is_file_suffix(filename, ('.nii.gz', '.nii', '.img.gz', '.img', '.hdr'), check_exist=check_exist)


def is_dicom(filename, check_exist=True):
    return is_file_suffix(filename, ('.dcm', '.dicom', '.DCM', '.DICOM'), check_exist=check_exist)


def make_empty_dir(dir_path, parents=False):
    """Make an empty directory. If it exists - check that it is empty"""
    dir_path = Path(dir_path)
    try:
        dir_path.mkdir(parents=parents, exist_ok=False)
    except FileExistsError:
        # the directory exists
        try:
            next(dir_path.glob('*'))
        except StopIteration:
            pass  # the directory exists but empty - ok
        else:
            raise FileExistsError(f'The directory "{dir_path}" is not empty')



import itk


def set_image_direction(image, direction):
    dim = 3
    direction_vnl_mat = itk.vnl_matrix_from_array(direction.astype('float').copy())  # copy is crucial for the float
    direction_itk = itk.Matrix[itk.D, dim, dim](direction_vnl_mat)
    image.SetDirection(direction_itk)


def get_image_direction(image):
    return itk.array_from_vnl_matrix(image.GetDirection().GetVnlMatrix().as_matrix())



def explicit_inds(key, shape):
    """Make getitem key explicit in the context of numpy ndarrays basic slicing and indexing"""
    ndim = len(shape)
    # set defaults
    start = [0] * ndim
    stop = list(shape)
    stride = [1] * ndim

    def update(i, k):
        """Update start, stop, stride at index i based on k"""
        if isinstance(k, int):
            start[i], stop[i] = k, k + 1
        elif isinstance(k, slice):
            start[i], stop[i], stride[i] = k.indices(shape[i])
        else:
            raise NotImplementedError(f'The indexing key "{k}" is not supported.')

    i = 0
    for k in key:
        if k is Ellipsis:
            # loop in reverse order
            i = ndim - 1
            for k_r in key[::-1]:
                if k_r is Ellipsis:
                    break
                else:
                    update(i, k_r)
                    i -= 1
            break
        else:
            update(i, k)
            i += 1

    return start, stop, stride



from pathlib import Path
from typing import Union

import itk
import numpy as np
from itkTemplate import TemplateTypeError

from medio.metadata.affine import Affine
from medio.metadata.dcm_uid import generate_uid
from medio.metadata.itk_orientation import itk_orientation_code
from medio.metadata.metadata import MetaData, check_dcm_ornt
from medio.utils.files import is_dicom, make_empty_dir


class ItkIO:
    coord_sys = 'itk'
    DEFAULT_COMPONENTS_AXIS = 0  # in the transposed image
    # default image type:
    dimension = 3
    pixel_type = itk.ctype('short')  # signed short - int16
    image_type = itk.Image[pixel_type, dimension]

    @staticmethod
    def read_img(input_path, desired_axcodes=None, header=False, components_axis=None, pixel_type=pixel_type,
                 fallback_only=True):
        """
        The main reader function, reads images and performs reorientation and unpacking
        :param input_path: path of image file or directory containing dicom series
        :param desired_axcodes: string or tuple - e.g. 'LPI', ('R', 'A', 'S')
        :param pixel_type: preferred itk pixel type for the image
        :param fallback_only: if True, finds the pixel_type automatically and uses pixel_type only if failed
        :param header: whether to include a header attribute with additional metadata in the returned metadata
        :param components_axis: if not None and the image is channeled (e.g. RGB) move the channels to channels_axis
        :return: numpy image and metadata object which includes pixdim, affine, original orientation string and
        coordinates system
        """
        input_path = Path(input_path)
        if input_path.is_dir():
            img = ItkIO.read_dir(str(input_path), pixel_type, fallback_only)
        elif input_path.is_file():
            img = ItkIO.read_img_file(str(input_path), pixel_type, fallback_only)
        else:
            raise FileNotFoundError(f'No such file or directory: "{input_path}"')

        affine = ItkIO.get_img_aff(img)
        metadata = MetaData(affine=affine, coord_sys=ItkIO.coord_sys)
        if (desired_axcodes is None) or (desired_axcodes == metadata.ornt):
            image_np = ItkIO.itk_img_to_array(img)
        else:
            orig_ornt = metadata.ornt  # store the original orientation before the reorientation
            img, _ = ItkIO.reorient(img, desired_axcodes)
            image_np, affine = ItkIO.unpack_img(img)
            metadata = MetaData(affine=affine, orig_ornt=orig_ornt, coord_sys=ItkIO.coord_sys)
        if header:
            metadict = img.GetMetaDataDictionary()
            metadata.header = {key: metadict[key] for key in metadict.GetKeys()}
        # TODO: consider unifying with PdcmIO.move_channels_axis
        n_components = img.GetNumberOfComponentsPerPixel()
        if (n_components > 1) and (components_axis is not None):
            # assert image_np.shape[ItkIO.DEFAULT_COMPONENTS_AXIS] == n_components  # TODO: is that always True?
            image_np = np.moveaxis(image_np, ItkIO.DEFAULT_COMPONENTS_AXIS, components_axis)
        return image_np, metadata

    @staticmethod
    def save_img(filename, image_np, metadata, use_original_ornt=True, components_axis=None,
                 allow_dcm_reorient=False, compression=False):
        """
        Save an image file with itk
        :param filename: the filename to save, str or os.PathLike
        :param image_np: the image's numpy array
        :param metadata: the corresponding metadata
        :param use_original_ornt: whether to save in the original orientation or not
        :param components_axis: if not None - the image has more than 1 component (e.g. RGB) and the components are in
        components_axis
        :param allow_dcm_reorient: whether to allow automatic reorientation to a right handed orientation or not
        :param compression: use compression or not
        """
        is_dcm = is_dicom(filename, check_exist=False)
        if is_dcm:
            image_np = ItkIO.prepare_dcm_array(image_np, is_vector=components_axis is not None)
        image = ItkIO.prepare_image(image_np, metadata, use_original_ornt, components_axis=components_axis,
                                    is_dcm=is_dcm, allow_dcm_reorient=allow_dcm_reorient)
        ItkIO.save_img_file(image, str(filename), compression=compression)

    @staticmethod
    def prepare_image(image_np, metadata, use_original_ornt, components_axis=None, is_dcm=False,
                      allow_dcm_reorient=False):
        """Prepare image for saving"""
        orig_coord_sys = metadata.coord_sys
        metadata.convert(ItkIO.coord_sys)
        desired_ornt = metadata.orig_ornt if use_original_ornt else None
        if is_dcm:
            # checking right-handed orientation before saving a dicom file/series
            desired_ornt = check_dcm_ornt(desired_ornt, metadata, allow_dcm_reorient=allow_dcm_reorient)
        image = ItkIO.pack2img(image_np, metadata.affine, components_axis=components_axis)
        if (desired_ornt is not None) and (desired_ornt != metadata.ornt):
            image, _ = ItkIO.reorient(image, desired_ornt)
        metadata.convert(orig_coord_sys)
        return image

    @staticmethod
    def prepare_dcm_array(image_np, is_vector=False):
        """Change image_np to correct data type for saving a single dicom file"""
        if is_vector:
            dcm_dtypes = [np.uint8]
        else:
            # for 3d image the supported data types are:
            dcm_dtypes = [np.uint8, np.uint16]
            # if the image is 2d it can be signed
            image_np_sq = np.squeeze(image_np)
            if image_np_sq.ndim == 2:
                # image_np = image_np_sq[..., np.newaxis]  # TODO: why?
                dcm_dtypes = [np.int16] + dcm_dtypes

        if image_np.dtype in dcm_dtypes:
            return image_np

        for dtype in dcm_dtypes:
            arr = image_np.astype(dtype, copy=False)
            if np.array_equal(arr, image_np):
                return arr

        raise NotImplementedError('Saving a single dicom file with ItkIO is currently supported only for \n'
                                  '1. 2d images - int16, uint16, uint8\n'
                                  '2. 3d images with integer nonnegative values - uint8, uint16\n'
                                  '3. 2d/3d RGB[A] images - uint8 (with channels_axis)\n'
                                  'For negative values, try to save a dicom directory or use PdcmIO.save_arr2dcm_file')

    @staticmethod
    def read_img_file(filename, pixel_type=None, fallback_only=False):
        """Common pixel types: itk.SS (int16), itk.US (uint16), itk.UC (uint8)"""
        return itk_imread(filename, pixel_type, fallback_only)

    @staticmethod
    def read_img_file_long(filename, image_type=image_type):
        """Longer version of read_img_file which returns the itk image and io engine string"""
        reader = itk.ImageFileReader[image_type].New()
        reader.SetFileName(filename)
        reader.Update()
        image_io = str(reader.GetImageIO()).split(' ')[0]
        image = reader.GetOutput()
        return image, image_io

    @staticmethod
    def save_img_file(image, filename, compression=False):
        itk.imwrite(image, filename, compression)

    @staticmethod
    def save_img_file_long(image, filename, compression=False):
        image_type = type(image)
        writer = itk.ImageFileWriter[image_type].New()
        if compression:
            writer.UseCompressionOn()
        writer.UseInputMetaDataDictionaryOn()
        writer.SetFileName(filename)
        writer.SetInput(image)
        writer.Update()

    @staticmethod
    def itk_img_to_array(img_itk):
        """Swap the axes to the usual x, y, z convention in RAI orientation (originally z, y, x)"""
        img_array = itk.array_from_image(img_itk).T  # the transpose here is equivalent to keep_axes=True
        return img_array

    @staticmethod
    def array_to_itk_img(img_array, components_axis=None):
        """Set components_axis to not None for vector images, e.g. RGB"""
        is_vector = False
        if components_axis is not None:
            img_array = np.moveaxis(img_array, components_axis, ItkIO.DEFAULT_COMPONENTS_AXIS)
            is_vector = True
        img_itk = itk.image_from_array(img_array.T.copy(), is_vector=is_vector)  # copy is crucial for the ordering
        return img_itk

    @staticmethod
    def unpack_img(img):
        image_np = ItkIO.itk_img_to_array(img)
        affine = ItkIO.get_img_aff(img)
        return image_np, affine

    @staticmethod
    def get_img_aff(img):
        direction = itk.array_from_vnl_matrix(img.GetDirection().GetVnlMatrix().as_matrix())
        spacing = itk.array_from_vnl_vector(img.GetSpacing().GetVnlVector())
        origin = itk.array_from_vnl_vector(img.GetOrigin().GetVnlVector())
        return Affine(direction=direction, spacing=spacing, origin=origin)

    @staticmethod
    def pack2img(image_np, affine, components_axis=None):
        image = ItkIO.array_to_itk_img(image_np, components_axis)
        ItkIO.set_img_aff(image, affine)
        return image

    @staticmethod
    def set_img_aff(image, affine):
        if not isinstance(affine, Affine):
            affine = Affine(affine)

        dimension = image.GetImageDimension()
        direction_arr, spacing, origin = affine.direction, affine.spacing, affine.origin

        # setting metadata
        spacing_vec = itk.Vector[itk.D, dimension]()
        spacing_vec.SetVnlVector(itk.vnl_vector_from_array(spacing.astype('float')))
        image.SetSpacing(spacing_vec)

        image.SetOrigin(origin.astype('float'))

        direction_mat = itk.vnl_matrix_from_array(direction_arr.astype('float'))
        direction = itk.Matrix[itk.D, dimension, dimension](direction_mat)
        image.SetDirection(direction)

    @staticmethod
    def reorient(img, desired_orientation: Union[int, tuple, str, None]):
        if desired_orientation is None:
            return img, None

        image_type = type(img)
        orient = itk.OrientImageFilter[image_type, image_type].New()
        orient.UseImageDirectionOn()
        orient.SetInput(img)
        if isinstance(desired_orientation, (str, tuple)):
            desired_orientation = itk_orientation_code(desired_orientation)
        orient.SetDesiredCoordinateOrientation(desired_orientation)
        orient.Update()
        reoriented_itk_img = orient.GetOutput()
        original_orientation_code = orient.GetGivenCoordinateOrientation()
        return reoriented_itk_img, original_orientation_code

    @staticmethod
    def read_dir(dirname, pixel_type=None, fallback_only=False):
        """
        Read a dicom directory. If there is more than one series in the directory an error is raised
        Shorter option for a single series (provided the slices order is known):
        >>> itk.imread([filename0, filename1, ...])
        """
        names_generator = itk.GDCMSeriesFileNames.New()
        names_generator.SetUseSeriesDetails(True)
        names_generator.AddSeriesRestriction('0008|0021')  # Series Date
        names_generator.SetGlobalWarningDisplay(False)
        names_generator.SetDirectory(dirname)

        series_uid = names_generator.GetSeriesUIDs()

        if len(series_uid) == 0:
            raise FileNotFoundError(f'No DICOMs in: "{dirname}"')
        if len(series_uid) > 1:
            raise OSError(f'The directory: "{dirname}"\n'
                          f'contains more than one DICOM series')

        series_identifier = series_uid[0]
        filenames = names_generator.GetFileNames(series_identifier)
        return itk_imread(filenames, pixel_type, fallback_only)

    @staticmethod
    def save_dcm_dir(dirname, image_np, metadata, use_original_ornt=True, components_axis=None, parents=False,
                     allow_dcm_reorient=False, **kwargs):
        """
        Save a 3d numpy array image_np as a dicom series of 2d dicom slices in the directory dirname
        :param dirname: the directory to save in the files, str or pathlib.Path. If it exists - must be empty
        :param image_np: the image's numpy array
        :param metadata: the corresponding metadata
        :param use_original_ornt: whether to save in the original orientation or not
        :param components_axis: if not None - the image has more than 1 component (e.g. RGB) and the components are in
        components_axis
        :param parents: if True, creates also the parents of dirname
        :param allow_dcm_reorient: whether to allow automatic reorientation to a right-handed orientation or not
        :param kwargs: optional kwargs passed to ItkIO.dcm_metadata: pattern, metadata_dict
        """
        image = ItkIO.prepare_image(image_np, metadata, use_original_ornt, components_axis=components_axis,
                                    is_dcm=True, allow_dcm_reorient=allow_dcm_reorient)
        image_type = type(image)
        _, (pixel_type, _) = itk.template(image)
        image2d_type = itk.Image[pixel_type, 2]
        writer = itk.ImageSeriesWriter[image_type, image2d_type].New()
        make_empty_dir(dirname, parents)
        # Generate necessary metadata and filenames per slice:
        mdict_list, filenames = ItkIO.dcm_series_metadata(image, dirname, **kwargs)
        metadict_vec = itk.vector[itk.MetaDataDictionary](mdict_list)
        writer.SetMetaDataDictionaryArray(metadict_vec)
        writer.SetFileNames(filenames)
        dicom_io = itk.GDCMImageIO.New()
        dicom_io.KeepOriginalUIDOn()
        writer.SetImageIO(dicom_io)
        writer.SetInput(image)
        writer.Update()

    @staticmethod
    def dcm_series_metadata(image, dirname, pattern='IM{}.dcm', metadata_dict=None):
        """
        Return dicom series metadata per slice and filenames
        :param image: the full itk image to be saved as dicom series
        :param dirname: the directory name
        :param pattern: str pattern for the filenames to save, including a placeholder ('{}') for the slice number
        :param metadata_dict: dictionary of metadata for adding tags or overriding the default values. For example,
        metadata_dict={'0008|0060': 'US'} will override the default 'CT' modality and set it to 'US' (ultrasound)
        :return: metadata dictionaries per slice, slice filenames
        """
        # The number of slices
        n = image.GetLargestPossibleRegion().GetSize().GetElement(2)

        # Shared properties for all the n slices:
        mdict = itk.MetaDataDictionary()

        # Series Instance UID
        mdict['0020|000e'] = generate_uid()
        # Study Instance UID
        mdict['0020|000d'] = generate_uid()

        # Pixel Spacing - TODO: not necessary - automatically saved
        spacing = image.GetSpacing()
        mdict['0028|0030'] = f'{spacing[0]}\\{spacing[1]}'
        # Spacing Between Slices
        mdict['0018|0088'] = str(spacing[2])
        # Image Orientation (Patient)
        orientation_str = '\\'.join([str(image.GetDirection().GetVnlMatrix().get(i, j))
                                     for j in range(2) for i in range(3)])
        mdict['0020|0037'] = orientation_str

        # Number of Frames
        mdict['0028|0008'] = '1'
        # Number of Slices
        mdict['0054|0081'] = str(n)
        # Modality
        mdict['0008|0060'] = 'CT'

        if metadata_dict is not None:
            for key, val in metadata_dict.items():
                mdict[key] = val

        # Per slice properties:
        mdict_list = []
        filenames = []

        for i in range(n):
            # copy the shared properties dict:
            mdict_i = itk.MetaDataDictionary(mdict)
            # Instance Number
            mdict_i['0020|0013'] = str(i + 1)
            # Image Position (Patient)
            position = image.TransformIndexToPhysicalPoint([0, 0, i])
            position_str = '\\'.join([str(position[i]) for i in range(3)])
            mdict_i['0020|0032'] = position_str

            mdict_list += [mdict_i]
            filenames += [str(Path(dirname) / pattern.format(i + 1))]

        return mdict_list, filenames


# TODO: remove imread below when itk.imread is updated in itk release
def itk_imread(filename, pixel_type=None, fallback_only=False):
    """Read an image from a file or series of files and return an itk.Image.
    The reader is instantiated with the image type of the image file if
    `pixel_type` is not provided (default). The dimension of the image is
    automatically found. If the given filename is a list or a tuple, the
    reader will use an itk.ImageSeriesReader object to read the files.
    If `fallback_only` is set to `True`, `imread()` will first try to
    automatically deduce the image pixel_type, and only use the given
    `pixel_type` if automatic deduction fails. Failures typically
    happen if the pixel type is not supported (e.g. it is not currently
    wrapped).

    This function is adapted from:
    ITK/Wrapping/Generators/Python/itkExtras.py
    The change from there is to accept also TemplateTypeError, making the flag fallback_only=True a legitimate default
    together with some default pixel type, e.g. pixel_type=itk.SS
    """
    # import itk  # originally in the itk - commented out here
    if fallback_only:
        if pixel_type is None:
            raise Exception("pixel_type must be set when using the fallback_only option")
        try:
            return itk_imread(filename)
        except (KeyError, TemplateTypeError):  # the change from itk's source, originally: `except KeyError:`
            pass
    if type(filename) in [list, tuple]:
        TemplateReaderType = itk.ImageSeriesReader
        io_filename = filename[0]
        increase_dimension = True
        kwargs = {'FileNames': filename}
    else:
        TemplateReaderType = itk.ImageFileReader
        io_filename = filename
        increase_dimension = False
        kwargs = {'FileName': filename}
    if pixel_type:
        # imageIO = itk.ImageIOFactory.CreateImageIO(io_filename, itk.CommonEnums.IOFileMode_ReadMode)
        read_mode = 0
        imageIO = itk.ImageIOFactory.CreateImageIO(io_filename, read_mode)
        if not imageIO:
            raise RuntimeError("No ImageIO is registered to handle the given file.")
        imageIO.SetFileName(io_filename)
        imageIO.ReadImageInformation()
        dimension = imageIO.GetNumberOfDimensions()
        # Increase dimension if last dimension is not of size one.
        if increase_dimension and imageIO.GetDimensions(dimension - 1) != 1:
            dimension += 1
        ImageType = itk.Image[pixel_type, dimension]
        reader = TemplateReaderType[ImageType].New(**kwargs)
        # TODO: consider
        #  reader.ForceOrthogonalDirectionOff()
        #  when 'FileNames' in kwargs
    else:
        reader = TemplateReaderType.New(**kwargs)
    reader.Update()
    return reader.GetOutput()



import itertools

import numpy as np

letter_vec_dict = {'R': [1, 0, 0],
                   'L': [-1, 0, 0],
                   'A': [0, 1, 0],
                   'P': [0, -1, 0],
                   'I': [0, 0, 1],
                   'S': [0, 0, -1]
                   }


def ornt2direction(ornt):
    return np.array([letter_vec_dict[letter] for letter in ornt]).T


ornt_iter = itertools.chain(
    *map(itertools.permutations,
         itertools.product(('R', 'L'), ('A', 'P'), ('I', 'S'))
         ))

# dictionary that translates itk orientation codes to direction matrices
ornt_direction_dict = dict()
ornt_list = []

for ornt_tup in ornt_iter:
    ornt = ''.join(ornt_tup)
    ornt_list += [ornt]
    ornt_direction_dict[ornt] = ornt2direction(ornt)


def is_right_handed_ornt(ornt):
    direction = ornt_direction_dict[ornt]
    det = np.linalg.det(direction)
    assert det in [-1, 1]  # check that the direction matrix is valid
    return det == 1


def direction2ornt(direction):
    for key, val in ornt_direction_dict.items():
        if np.array_equal(direction, val):
            return key
    raise ValueError('Invalid direction')



from setuptools import setup, find_packages

from medio import __version__

with open('README.md') as f:
    long_description = f.read()


setup(name='medio',
      version=__version__,
      description='Medical images I/O python package',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/RSIP-Vision/medio',
      author='RSIP Vision',
      author_email='info@rsipvision.com',
      keywords=['medical-images', 'IO', 'itk', 'nibabel', 'pydicom', 'python'],
      packages=find_packages(exclude=['*.tests']),
      install_requires=[
            'itk-io == 5.0.1',
            'itk-filtering == 5.0.1',
            'nibabel >= 3.0.2',
            'pydicom >= 2.0.0',
            'dicom-numpy >= 0.4.0',
            'numpy >= 1.18.1'
      ],
      python_requires='>=3.6',
      classifiers=[
            'Development Status :: 3 - Alpha',

            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'Intended Audience :: Healthcare Industry',
            'Intended Audience :: Science/Research',

            'Topic :: Scientific/Engineering :: Medical Science Apps.',
            'Topic :: Software Development :: Libraries :: Python Modules',

            'Programming Language :: Python',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',

            'Operating System :: OS Independent'
      ],
      license='Apache License 2.0',
      zip_safe=True)



from medio.metadata.affine import Affine
from medio.metadata.metadata import MetaData
from medio.read_save import read_img, save_img
from medio.utils.explicit_slicing import explicit_inds


class MedImg:
    def __init__(self, np_image, metadata, filename=None, **kwargs):
        """
        Class for a single medical image, represented by numpy array and metadata object - image affine, original
        orientation and coordinate system. The class allows performing operations on the image with respective update of
        the metadata (mainly affine).
        The class can be instantiated also with a filename:
        >>> MedImg(None, None, '../scan1.mhd')
        With possible kwargs for the reader.
        """
        if filename is not None:
            np_image, metadata = read_img(filename, **kwargs)
        self.np_image = np_image
        self.metadata = metadata

    def save(self, filename, **kwargs):
        save_img(filename, self.np_image, self.metadata, **kwargs)

    def __getitem__(self, item):
        """
        This method allows cropping and basic down-sampling:
        >>> mimg = MedImg(np_image, metadata)
        >>> new_mimg = mimg[:, 4:-4, ::3]
        >>> print(new_mimg.metadata)
        Ellipsis (...) is also supported
        """
        np_image = self.np_image[item]
        start, stop, stride = explicit_inds(item, self.np_image.shape)
        affine = Affine(self.metadata.Affine.copy())
        affine.origin = affine.index2coord(start)
        affine.spacing = affine.spacing * stride
        metadata = MetaData(affine, self.metadata.orig_ornt, self.metadata.coord_sys)
        return MedImg(np_image, metadata)



from pathlib import Path

from medio.backends.itk_io import ItkIO
from medio.backends.nib_io import NibIO
from medio.backends.pdcm_io import PdcmIO
from medio.metadata.convert_nib_itk import inv_axcodes
from medio.utils.files import is_nifti


def read_img(input_path, desired_ornt=None, backend=None, dtype=None, header=False, channels_axis=-1, **kwargs):
    """
    Read medical image with nibabel or itk
    :param input_path: str or os.PathLike, the input path of image file or a directory containing dicom series
    :param desired_ornt: optional parameter for reorienting the image to desired orientation, e.g. 'RAS'
    The desired_ornt string is in itk standard (if NibIO is used, the orientation is converted accordingly)
    :param backend: optional parameter for setting the reader backend: 'itk', 'nib', 'pdcm' (also 'pydicom') or None
    :param dtype: equivalent to np_image.astype(dtype) if dtype is not None
    :param header: if True, the returned metadata will include a header attribute with additional metadata dictionary as
    read by the backend. Note: currently, this is supported for files only
    :param channels_axis: if not None and the image is channeled (e.g. RGB) move the channels to channels_axis in the
    returned image array
    :return: numpy image and metadata object
    """
    nib_reader = NibIO.read_img
    itk_reader = ItkIO.read_img
    pdcm_reader = PdcmIO.read_img
    if backend is None:
        if is_nifti(input_path):
            reader = nib_reader
        else:
            reader = itk_reader
    else:
        if backend == 'nib':
            reader = nib_reader
        elif backend == 'itk':
            reader = itk_reader
        elif backend in ('pdcm', 'pydicom'):
            if desired_ornt is not None:
                raise NotImplementedError('Pydicom reader backend does not yet support reorientation. The passed '
                                          'desired orientation must be None (default).')
            reader = pdcm_reader
        else:
            raise ValueError('The backend argument must be one of: "itk", "nib", "pdcm" (or "pydicom"), None')

    if reader == nib_reader:
        desired_ornt = inv_axcodes(desired_ornt)

    if reader is pdcm_reader:
        np_image, metadata = reader(input_path, header, channels_axis, **kwargs)
    else:
        np_image, metadata = reader(input_path, desired_ornt, header, channels_axis, **kwargs)

    if dtype is not None:
        np_image = np_image.astype(dtype, copy=False)
    return np_image, metadata


def save_img(filename, np_image, metadata, use_original_ornt=True, backend=None, dtype=None, channels_axis=None,
             mkdir=False, parents=False, **kwargs):
    """
    Save numpy image with corresponding metadata to file
    :param filename: str or os.PathLike, the output filename
    :param np_image: the numpy image
    :param metadata: the metadata of the image
    :param use_original_ornt: whether to save in the original orientation stored in metadata.orig_ornt or not
    :param backend: optional - 'itk', 'nib' or None
    :param dtype: equivalent to np_image.astype(dtype) if dtype is not None
    :param channels_axis: if not None - the image is channeled (e.g. RGB) and the channels are in channels_axis
    :param mkdir: if True, creates the directory of `filename`
    :param parents: to be used with `mkdir=True`. If True, creates also the parent directories
    """
    nib_writer = NibIO.save_img
    itk_writer = ItkIO.save_img
    if backend is None:
        if is_nifti(filename, check_exist=False):
            writer = nib_writer
        else:
            writer = itk_writer
    else:
        if backend == 'nib':
            writer = nib_writer
        elif backend == 'itk':
            writer = itk_writer
        else:
            raise ValueError('The backend argument must be one of: "itk", "nib", None')
    if mkdir:
        Path(filename).parent.mkdir(parents=parents, exist_ok=True)
    if dtype is not None:
        np_image = np_image.astype(dtype, copy=False)
    writer(filename, np_image, metadata, use_original_ornt, channels_axis, **kwargs)


def save_dir(dirname, np_image, metadata, use_original_ornt=True, dtype=None, channels_axis=None, parents=False,
             allow_dcm_reorient=False, **kwargs):
    """Save image as a dicom directory. See medio.backends.itk_io.ItkIO.save_dcm_dir documentation.
    dtype is equivalent to passing image_np.astype(dtype) if dtype is not None"""
    if dtype is not None:
        np_image = np_image.astype(dtype, copy=False)
    ItkIO.save_dcm_dir(dirname, np_image, metadata, use_original_ornt, channels_axis, parents, allow_dcm_reorient,
                       **kwargs)



from pydicom.dataset import FileDataset


def convert_ds(dataset):
    """
    Convert a dicom file with shared functional groups sequence into a class which enables shorter access to its
    properties. This is intended primarily for a single dicom file which includes several frames/slices.
    Usage example:
    >>> import pydicom
    >>> ds = pydicom.dcmread('single_dicom_file/I0')
    >>> ds = convert_ds(ds)
    >>> print(ds.PixelSpacing)
    """
    if dataset.get('NumberOfFrames', 1) > 1:
        dataset.__class__ = MultiFrameFileDataset
    return dataset


class MultiFrameFileDataset(FileDataset):
    """This class enables shorter access to basic properties of pydicom dataset of a certain type"""
    @property
    def ImageOrientationPatient(self):
        return self.SharedFunctionalGroupsSequence[0].PlaneOrientationSequence[0].ImageOrientationPatient

    @property
    def PixelSpacing(self):
        return self.SharedFunctionalGroupsSequence[0].PixelMeasuresSequence[0].PixelSpacing

    @property
    def SpacingBetweenSlices(self):
        return self.SharedFunctionalGroupsSequence[0].PixelMeasuresSequence[0].SpacingBetweenSlices

    @property
    def SliceThickness(self):
        return self.SharedFunctionalGroupsSequence[0].PixelMeasuresSequence[0].SliceThickness

    @property
    def RescaleIntercept(self):
        return self.SharedFunctionalGroupsSequence[0].PixelValueTransformationSequence[0].RescaleIntercept

    @property
    def RescaleSlope(self):
        return self.SharedFunctionalGroupsSequence[0].PixelValueTransformationSequence[0].RescaleSlope

    @property
    def ImagePositionPatient(self):
        """Note: this property returns only the position of the first slice"""
        return self.PerFrameFunctionalGroupsSequence[0].PlanePositionSequence[0].ImagePositionPatient

    def slice_positions(self):
        """Return a list of the slices' position"""
        return [seq.PlanePositionSequence[0].ImagePositionPatient for seq in self.PerFrameFunctionalGroupsSequence]

    def slice_position(self, index):
        """Return the slice position according to the slice index"""
        return self.PerFrameFunctionalGroupsSequence[index].PlanePositionSequence[0].ImagePositionPatient

    def del_intensity_trans(self):
        """Delete the pixel value transformation sequence from dataset"""
        del self.SharedFunctionalGroupsSequence[0].PixelValueTransformationSequence



from medio.metadata.affine import Affine
from medio.metadata.metadata import MetaData
from medio.read_save import read_img, save_img, save_dir
from medio import backends, metadata, medimg, utils

__version__ = '0.3.0'

__all__ = ['read_img', 'save_img', 'save_dir', 'MetaData', 'Affine', '__version__']
